#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include "hashtab.h"

#define HOP_LOG "/tmp/scratchpad/hop.txt"

// lower case char for unicast, upper case char for multicast
#define PRE_GEN    "LOG: g"    /* packet Generation */
#define PRE_MGEN   "LOG: G"    /* New multicast packet */
#define PRE_SEND   "LOG: s"    /* packet Send */
#define PRE_MCAST  "LOG: S"    /* send multicast packet */
#define PRE_DLG    "LOG: e"    /* dElegated packet */
#define PRE_EDLG   "LOG: E"    /* dElegated multicast packet */
#define PRE_DELAY  "LOG: d"    /* packet Delayed send of stored packet*/
#define PRE_MDELAY "LOG: D"    /* packet Delayed send of stored packet*/
#define PRE_STORE  "LOG: t"    /* packet sTored */
#define PRE_MSTORE "LOG: T"    /* packet sTored */
#define PRE_RECV   "LOG: r"    /* packet Received */
#define PRE_MRECV  "LOG: R"    /* Received multicast packet */
#define PRE_DUP    "LOG: i"    /* ignored duplicate unicast packet */
#define PRE_MDUP   "LOG: I"    /* ignored duplicate multicast packet */
#define PRE_ODROP  "LOG: o"    /* drop Out of interval multicast packet*/
#define PRE_BUFF   "LOG: b"    /* Buffer size at node */
#define PRE_CBUFF  "LOG: B"    /* Buffer size at core node */
#define PRE_GMBR   "LOG: M"    /* group membership */
#define PRE_PDEL   "LOG: p"    /* group membership */

#define MAXLINE 64
#define MINLINE 8

#define DEBUG 1

extern int errno;

int main(int argc, char *argv[])
{
   char line[MAXLINE];
   char *cp;
   FILE *fp ;
   FILE *hfp;
   int len;

   int   src;
   int   psrc;
   int   dest;
   int   gid;
   int   seqno;
   float time;     // current time
   float gtime;    // time message generated
   int   bufsize;
   int   hop;
   float buf_occupancy;

   float total_bufsize = 0.0;
   float total_occupancy = 0.0;
   float total_cbufsize = 0.0;
   float total_coccupancy = 0.0;
   int   total_hop = 0;

   int no_gen = 0;
   int no_send = 0;
   int no_delay = 0;
   int no_recv = 0;
   int no_store = 0;
   int no_buff  = 0;
   int no_cbuff = 0;
   int no_hop = 0;

   int no_mgen  = 0; // number of multicast packets generated
   int no_msend = 0; // number of multicast packets sent 
   int no_mrecv = 0; // number of multicast packets received
   int no_mdup  = 0; // number of multicast packets ignored as duplicates

   float mdr = 0.0 ; // message delivery ratio
   float pdelay = 0.0 ; // percentage of delayed packets
   //float psend  = 0.0 ; // percentage of packets sent 
   float pstore  = 0.0 ; // percentage of packets sent 

   float mmdr = 0.0 ; // multicast message delivery ratio

   char str[48];

   if ( argc == 1 ) {
      fp = stdin;
   } else if ( argc == 2 ) {
      fp = (FILE *) fopen(argv[1], "r");
   } else {
      fprintf(stderr, "usage: %s logfile\n", argv[0]);
      exit(1);
   }

   if ( fp == NULL ) {
      fprintf(stderr, "Could not open input file: %s\n", argc == 2 ? argv[1] : "stdin");
      fprintf(stderr, "errno: %d (%s)\n", errno, strerror(errno));
      exit(1);
   }

   hfp = fopen(HOP_LOG, "w");
   if ( hfp == NULL ) {
      fprintf(stderr, "Could not open hop count log file: %s, skipping.\n", HOP_LOG);
   }

   memset(line, 0, sizeof(line));

   init();

   // read the logs
   while ( fgets(line, sizeof(line), fp) != NULL ) {
      len = strlen(line);
      if ( len < MINLINE ) {
         continue; // not a valid log line
      }

      // get rid of newline character
      if ( line[len-1] == '\n' ) {
          line[len-1] = '\0';
      }

      if ( strncmp(line, PRE_GEN, sizeof(PRE_GEN)-1) == 0 ) {
         no_gen++;

         //PRE_GEN format: g _%d_ _%d_ [%d] %g {Format: g _src_ _dest_ [seqno] time}
         cp = line + sizeof(PRE_GEN);
         sscanf(cp, "_%d_ _%d_ [%d] %f", &src, &dest, &seqno, &time);
         //printf ( "gen: src: %d, dest: %d, sno: %d, time: %f\n", src, dest, seqno, time);

         memset(str, 0, sizeof(str));
         snprintf(str, sizeof(str), "%d:%d:%d", src, dest, seqno);
         lookup ( LOG_GEN, str, time, TRUE ) ;  
      } else if ( strncmp(line, PRE_SEND, sizeof(PRE_SEND)-1) == 0 ) {
         no_send++;

         //PRE_SEND format: s _%d_ _%d_ [%d] %g {Format: s _src_ _dest_ [seqno] time}
         cp = line + sizeof(PRE_SEND);
         sscanf(cp, "_%d_ _%d_ [%d] %f", &src, &dest, &seqno, &time);
         //printf ( "send: src: %d, dest: %d, sno: %d, time: %f\n", src, dest, seqno, time);
      } else if ( strncmp(line, PRE_DELAY, sizeof(PRE_DELAY)-1) == 0 ) {
         no_delay++;

         //PRE_DELAY format: d _%d_ _%d_ [%d] %g {Format: d _src_ _dest_ [seqno] time}
         cp = line + sizeof(PRE_DELAY);
         sscanf(cp, "_%d_ _%d_ [%d] %f", &src, &dest, &seqno, &time);
         //printf ( "delay: src: %d, dest: %d, sno: %d, time: %f\n", src, dest, seqno, time);
      } else if ( strncmp(line, PRE_STORE, sizeof(PRE_STORE)-1) == 0 ) {
         no_store++;

         cp = line + sizeof(PRE_STORE);
         //PRE_STORE format: t _%d_ _%d_ [%d] %g {Format: t _src_ _dest_ [gid] [seqno] time}
         sscanf(cp, "_%d_ _%d_ [%d] [%d] %f", &src, &dest, &gid, &seqno, &time);
         //printf ( "store: src: %d, dest: %d, sno: %d, time: %f\n", src, dest, seqno, time);
      } else if ( strncmp(line, PRE_BUFF, sizeof(PRE_BUFF)-1) == 0 ) {
         no_buff++;

         cp = line + sizeof(PRE_BUFF);
         //PRE_BUFF format: b _%d_ _%d_ %g {Format: b _src_ buffer_size buffer_occupancy time}
         sscanf(cp, "_%d_ %d %f %f", &src, &bufsize, &buf_occupancy, &time);

         total_bufsize += bufsize;
         total_occupancy += buf_occupancy;
      } else if ( strncmp(line, PRE_CBUFF, sizeof(PRE_CBUFF)-1) == 0 ) {
         no_cbuff++;

         cp = line + sizeof(PRE_CBUFF);
         //PRE_CBUFF format: B _%d_ _%d_ %g {Format: b _src_ buffer_size buffer_occupancy time}
         sscanf(cp, "_%d_ %d %f %f", &src, &bufsize, &buf_occupancy, &time);

         total_cbufsize += bufsize;
         total_coccupancy += buf_occupancy;
      } else if ( strncmp(line, PRE_RECV, sizeof(PRE_RECV)-1) == 0 ) {
         cp = line + sizeof(PRE_RECV);
         //PRE_RECV format: r _%d_ _%d_ [%d] %g {Format: r _dest_ _src_ psrc [seqno] gen_time curr_time hop}
         sscanf(cp, "_%d_ _%d_ %d [%d] %f %f %d", &dest, &src, &psrc, &seqno, &gtime, &time, &hop);
         //printf ( "recv: src: %d, dest: %d, sno: %d, time: %f\n", src, dest, seqno, time);

         memset(str, 0, sizeof(str));
         snprintf(str, sizeof(str), "%d:%d:%d", src, dest, seqno);
         if ( lookup ( LOG_RECV, str, time, FALSE ) == NULL ) {
            no_recv++;
            no_hop++;
            total_hop += hop;
            if ( hfp != NULL ) {
               fprintf(hfp, "%d\n", hop);
            }
         } else {
            //printf ( "duplicate entry: %s\t %f\n", str, time);
         }
         lookup ( LOG_RECV, str, time, TRUE ) ;  
      } else if ( strncmp(line, PRE_MGEN, sizeof(PRE_MGEN)-1) == 0 ) {
         no_mgen++;

         //PRE_MGEN format: G _%d_ [%d] [%d] %g {Format: G _src_ [seqno] [gid] time}
         cp = line + sizeof(PRE_MGEN);
         sscanf(cp, "_%d_ [%d] [%d] %f", &src, &seqno, &gid, &time);

         memset(str, 0, sizeof(str));
         snprintf(str, sizeof(str), "%d:%d:%d", src, seqno, gid);
         lookup(LOG_MGEN, str, time, TRUE);  
      } else if ( strncmp(line, PRE_MRECV, sizeof(PRE_MRECV)-1) == 0 ) {
         cp = line + sizeof(PRE_MRECV);

         //PRE_MRECV format: R _%d_ _%d_ [%d] [%d] %g {Format: R _dest_ _src_ psrc [seqno] [gid] gen_time curr_time hop}
         sscanf(cp, "_%d_ _%d_ %d [%d] [%d] %f %f %d", &dest, &src, &psrc, &seqno, &gid, &gtime, &time, &hop);

         memset(str, 0, sizeof(str));
         snprintf(str, sizeof(str), "%d:%d:%d", src, dest, seqno);
         if ( lookup ( LOG_MRECV, str, time, FALSE ) == NULL ) {
            no_mrecv++;
            total_hop += hop;
            no_hop++;
            if ( hfp != NULL ) {
               fprintf(hfp, "%d\n", hop);
            }
         } else {
            //printf ( "duplicate entry: %s\t %f\n", str, time);
         }
         lookup ( LOG_MRECV, str, time, TRUE ) ;  
      } else if ( strncmp(line, PRE_GMBR, sizeof(PRE_GMBR)-1) == 0 ) {
         cp = line + sizeof(PRE_GMBR);

         //PRE_GMBR format: M _%d_ [%d] %g {Format: M _src_ [gid] time}
         sscanf(cp, "_%d_ [%d] %f", &src, &gid, &time);

         grp_lookup(gid, src, time, TRUE);
      } else if ( strncmp(line, PRE_MCAST, sizeof(PRE_MCAST)-1) == 0 ) {
         no_msend++;
      } else if ( strncmp(line, PRE_PDEL, sizeof(PRE_PDEL)-1) == 0 ) {
         no_msend++;
      } else if ( strncmp(line, PRE_MDUP, sizeof(PRE_MDUP)-1) == 0 ) {
         no_mdup++;
      } else {
         if ( DEBUG ) {
            // ignore invalid log entries
            // fprintf(stderr, "unrecognized line: \"%s\"\n", line);
         }
      }
   }

   if ( hfp != NULL ) {
      fclose(hfp);
   }

   mdr = ((float) no_recv) / ((float) no_gen);
   pstore = ((float) no_store) / ((float) no_gen);
   pdelay = ((float) no_delay) / ((float) no_store);

   mmdr = ((float) no_mrecv) / ((float) no_mgen); // multicast message delivery ratio

   //printf ("\nhashtab contents (generated):\n");
   //print_all(LOG_GEN);

   //printf ("\nhashtab contents (received):\n");
   //print_all(LOG_RECV);

   printf ("\n------------------- statistics -------------------\n\n");

   printf ("\n-------- unicast --------\n\n");

   if ( no_gen > 0 ) {
      print_stats(FALSE);
      printf ("\n");

      printf ("No. messages generated:\t %d\n", no_gen);
      printf ("No. messages received:\t %d\n", no_recv);
      printf ("No. messages stored:\t %d\n", no_store);

      printf("\n");

      printf ("Message Delivery  Ratio:\t %f\n", mdr);
      printf ("Average buffer occupancy:\t %f %f\n", total_bufsize/no_buff, total_occupancy/((float)no_buff) );
      printf ("Average core buffer occupancy:\t %f %f\n", total_cbufsize/no_cbuff, total_coccupancy/((float)no_cbuff) );
      printf ("Average number of hops:\t %f\n", ((float) total_hop)/no_hop);
      printf ("Percent of stored  packets:\t %f\n", pstore);
      printf ("Percent of delayed packets:\t %f\n", pdelay);
      printf ("Percent of pkts left in buffer:\t %f (not sent)\n", 1.0-pdelay);
   } else {
      printf("No unicast packets found.\n");
   }

   printf("\n");

   printf ("\n-------- multicast --------\n\n");
   if ( no_mgen > 0 ) {
      print_mstats(FALSE);

      printf ("No. messages generated  (MC):\t %d\n", no_mgen);
      printf ("No. messages sent (all) (MC):\t %d\n", no_msend);
      printf ("No. messages received   (MC):\t %d\n", no_mrecv);
      printf ("Average goodput ratio   (MC):\t %f\n", no_mrecv/((float)no_msend));
      printf ("No. messages ignored    (MC):\t %d\n", no_mdup);
      printf ("Average buffer occupancy(MC):\t %f %f\n", total_bufsize/no_buff, total_occupancy/((float)no_buff));
      printf ("Average core buffer occupancy(MC):\t %f %f\n", total_coccupancy/((float)no_cbuff), total_coccupancy);
      printf ("Average number of hops:\t %f\n", ((float) total_hop)/no_hop);
   } else {
      printf("No multicast packets found.\n");
   }

   printf("\n");

   return 0;
}
